@@ -80,7 +80,7 @@ use Geo::OGC::Service;
 use vars qw(@ISA);
 push @ISA, qw(Geo::OGC::Service::Common);
 
-our $VERSION = '0.02';
+our $VERSION = '0.03';
 
 our $radius_of_earth_at_equator = 6378137;
 our $standard_pixel_size = 0.28 / 1000;
@@ -186,7 +186,9 @@ sub GetCapabilities {
     $writer->close_element;
     $writer->open_element(Contents => {});
 
-    my $t_srs = Geo::OSR::SpatialReference->new(EPSG => 4326);
+    my $t_srs = $Geo::GDAL::VERSION >= 2 ? 
+        Geo::OSR::SpatialReference->new(EPSG=>4326) : 
+        Geo::OSR::SpatialReference->create(EPSG=>4326);
 
     for my $set (@{$self->{config}{TileSets}}) {
         my $projection = $projections{$set->{SRS}};
@@ -194,7 +196,9 @@ sub GetCapabilities {
         my $bb;
         if ($set->{BoundingBox}) {
             my ($epsg) = $set->{BoundingBox}{SRS} =~ /(\d+)/;
-            my $s_srs = Geo::OSR::SpatialReference->new(EPSG => $epsg);
+            my $s_srs = $Geo::GDAL::VERSION >= 2 ? 
+                Geo::OSR::SpatialReference->new(EPSG => $epsg) :
+                Geo::OSR::SpatialReference->create(EPSG => $epsg);
             my $ct = Geo::OSR::CoordinateTransformation->new($s_srs, $t_srs);
 
             my $x = $set->{BoundingBox};
@@ -332,7 +336,7 @@ sub WMSGetCapabilities {
 
 Serves the tile request if WMS is used.
 
-Sends the requested tile based on parameters BBOX and LAYERS.
+Sends the requested tile based on parameters BBOX, LAYERS, and SRS.
 
 The tiles should be in a tile map resource type of directory structure
 (z/y/x.png). The value of the 'path' key in the TileSet config element
@@ -342,7 +346,7 @@ should point to the directory.
 
 sub GetMap {
     my ($self) = @_;
-    for my $param (qw/bbox layers/) {
+    for my $param (qw/bbox layers srs/) {
         unless ($self->{parameters}{$param}) {
             $self->error({ exceptionCode => 'MissingParameterValue',
                            locator => uc($param) });
@@ -362,47 +366,48 @@ sub GetMap {
         return;
     }
 
-    my $projection = $projections{$set->{SRS}};
-    my @resolutions;
+    my $projection = $projections{$self->{parameters}{srs}};
+
+    unless ($projection) {
+        my @supported = sort keys %projections;
+        return $self->error({ exceptionCode => 'InvalidParameterValue',
+                              locator => 'SRS',
+                              ExceptionText => "$self->{parameters}{srs} is not currently supported." });
+    }
+
     my $extent_width = $projection->{extent}{maxx} - $projection->{extent}{minx};
-    for my $i (0..19) {
-        $resolutions[$i] = $extent_width/(2**$i * $tile_width);
+    my $extent_height = $projection->{extent}{maxy} - $projection->{extent}{miny};
+    my @bbox = split /,/, $self->{parameters}{bbox}; # minx, miny, maxx, maxy
+    my $width = $bbox[2] - $bbox[0];
+    my $height = $bbox[3] - $bbox[1];
+    my $matrix = 0;
+    my $two_to_matrix = 1;
+    while (abs($two_to_matrix * $width - $extent_width) > 10) {
+        ++$matrix;
+        $two_to_matrix *= 2;
     }
     
-    my @bbox = split /,/, $self->{parameters}{bbox}; # minx, miny, maxx, maxy
-    my $units_per_pixel = ($bbox[2]-$bbox[0])/$tile_width;
-    my $matrix;
-    my $res;
-    my $i = 0;
-    for my $r (@resolutions) {
-        if (abs($r - $units_per_pixel) < 0.1) {
-            $res = $r;
-            $matrix = $i;
-            last;
-        }
-        $i++;
+    my $col = $two_to_matrix * ($bbox[0] - $projection->{extent}{minx}) / $extent_width;
+    $col = int( POSIX::ceil($col) - 1);
+    my $row = $two_to_matrix * ($projection->{extent}{maxy} - $bbox[3]) / $extent_height;
+    $row = int( POSIX::ceil($row) - 1);
+
+    if ($self->{parameters}{srs} eq 'EPSG:3067') {
+        # I don't know why this is needed, but it is
+        $col++;
+        $row++;
     }
-    my $rows = 2**$matrix;
-
-    # from globalmaptiles.py by Klokan Petr Pridal:
-
-    my $originShift = 2 * pi * 6378137 / 2.0;
-
-    my $px = ($bbox[0] + $originShift) / $res;
-    my $py = ($bbox[1] + $originShift) / $res;
-
-    my $col = int( POSIX::ceil( $px / $tile_width ) - 1 );
-    my $row = int( POSIX::ceil( $py / $tile_height ) - 1 );
 
     ($set->{ext}) = $set->{Format} =~ /(\w+)$/;
 
     if ($set->{file}) {
         $self->{parameters}{tilematrix} = $matrix;
-        $self->{parameters}{tilecol} = $row;
-        $self->{parameters}{tilerow} = 2**$matrix-($col+1);
+        $self->{parameters}{tilecol} = $col;
+        $self->{parameters}{tilerow} = $row;
         return $self->make_tile($set);
     }
 
+    $row = $two_to_matrix - $row - 1;
     return $self->tile("$set->{path}/$matrix/$col/$row.$set->{ext}", $set->{Format});
 }
 
@@ -523,7 +528,7 @@ sub FeatureInfo {
 sub make_tile {
     my ($self, $set) = @_;
 
-    $self->log($self->{parameters});
+    #$self->log($self->{parameters});
 
     return $self->error({ exceptionCode => 'ResourceNotFound',
                           ExceptionText => "File resources are not supported by this GDAL version." })
@@ -535,7 +540,9 @@ sub make_tile {
         my $srs_s = $ds->SpatialReference;
         
         my ($epsg_t) = $set->{SRS} =~ /(\d+)/;
-        my $srs_t = Geo::OSR::SpatialReference->new(EPSG => $epsg_t);
+        my $srs_t = $Geo::GDAL::VERSION >= 2 ? 
+            Geo::OSR::SpatialReference->new(EPSG => $epsg_t) :
+            Geo::OSR::SpatialReference->create(EPSG => $epsg_t);
         
         if (!$srs_s->IsSame($srs_t)) {
             $ds = $ds->Warp('/vsimem/w.png', );
@@ -544,18 +551,20 @@ sub make_tile {
 
     my $projection = $projections{$set->{SRS}};
         
-    my $tile = Tile->new($projection->{extent}, $self->{parameters});
+    my $tile = Geo::OGC::Service::WMTS::Tile->new($projection->{extent}, $self->{parameters});
 
     eval {
    
         if ($set->{processing}) {
-            $tile->expand(1);
-            $ds = $ds->Translate( "/vsimem/tmp.png", ['-of' => 'PNG', '-r' => 'bilinear' , 
+            $tile->expand(2);
+            $ds = $ds->Translate( "/vsimem/tmp.tiff", ['-of' => 'GTiff', '-r' => 'bilinear' , 
                                                       '-outsize' , $tile->tile,
                                                       '-projwin', $tile->projwin] );
             my $z = $set->{zFactor} // 1;
-            $ds = $ds->DEMProcessing("/vsimem/tmp2.png", $set->{processing}, undef, { of => 'PNG', z => $z });
-            $tile->expand(-1);
+            $ds = $ds->DEMProcessing("/vsimem/tmp2.tiff", $set->{processing}, undef, { of => 'GTiff', z => $z });
+            $tile->expand(-2);
+        } elsif ($self->{processor}) {
+            $ds = $self->{processor}->process($ds, $tile);
         }
         
         my $writer = $self->{responder}->([200, [ 'Content-Type' => "image/png" ]]);
